"""
SQL
---
The SQL module provides a SQL-based version of the :py:class:`.DBContextStorage` class.
This class is used to store and retrieve context data from SQL databases.
It allows Chatsky to easily store and retrieve context data in a format that is highly scalable
and easy to work with.

The SQL module provides the ability to choose the backend of your choice from
MySQL, PostgreSQL, or SQLite. You can choose the one that is most suitable for your use case and environment.
MySQL and PostgreSQL are widely used open-source relational databases that are known for their
reliability and scalability. SQLite is a self-contained, high-reliability, embedded, full-featured,
public-domain, SQL database engine.
"""

import asyncio
import importlib
import json
from typing import Hashable

from chatsky.script import Context

from .database import DBContextStorage, threadsafe_method
from .protocol import get_protocol_install_suggestion

try:
    from sqlalchemy import Table, MetaData, Column, JSON, String, inspect, select, delete, func
    from sqlalchemy.ext.asyncio import create_async_engine

    sqlalchemy_available = True
except (ImportError, ModuleNotFoundError):
    sqlalchemy_available = False

postgres_available = sqlite_available = mysql_available = False

try:
    import asyncpg

    _ = asyncpg

    postgres_available = True
except (ImportError, ModuleNotFoundError):
    pass

try:
    import asyncmy

    _ = asyncmy

    mysql_available = True
except (ImportError, ModuleNotFoundError):
    pass

try:
    import aiosqlite

    _ = aiosqlite

    sqlite_available = True
except (ImportError, ModuleNotFoundError):
    pass

if not sqlalchemy_available:
    postgres_available = sqlite_available = mysql_available = False


def import_insert_for_dialect(dialect: str):
    """
    Imports the insert function into global scope depending on the chosen sqlalchemy dialect.

    :param dialect: Chosen sqlalchemy dialect.
    """
    global insert
    insert = getattr(
        importlib.import_module(f"sqlalchemy.dialects.{dialect}"),
        "insert",
    )


class SQLContextStorage(DBContextStorage):
    """
    | SQL-based version of the :py:class:`.DBContextStorage`.
    | Compatible with MySQL, Postgresql, Sqlite.

    :param path: Standard sqlalchemy URI string.
        When using sqlite backend in Windows, keep in mind that you have to use double backslashes '\\'
        instead of forward slashes '/' in the file path.
    :param table_name: The name of the table to use.
    :param custom_driver: If you intend to use some other database driver instead of the recommended ones,
        set this parameter to `True` to bypass the import checks.
    """

    def __init__(self, path: str, table_name: str = "contexts", custom_driver: bool = False):
        DBContextStorage.__init__(self, path)

        self._check_availability(custom_driver)
        self.engine = create_async_engine(self.full_path, pool_pre_ping=True)
        self.dialect: str = self.engine.dialect.name

        id_column_args = {"primary_key": True}
        if self.dialect == "sqlite":
            id_column_args["sqlite_on_conflict_primary_key"] = "REPLACE"

        self.metadata = MetaData()
        self.table = Table(
            table_name,
            self.metadata,
            Column("id", String(36), **id_column_args),
            Column("context", JSON),  # column for storing serialized contexts
        )

        asyncio.run(self._create_self_table())

        import_insert_for_dialect(self.dialect)

    @threadsafe_method
    async def set_item_async(self, key: Hashable, value: Context):
        value = value if isinstance(value, Context) else Context.cast(value)
        value = json.loads(value.model_dump_json())

        insert_stmt = insert(self.table).values(id=str(key), context=value)
        update_stmt = await self._get_update_stmt(insert_stmt)

        async with self.engine.connect() as conn:
            await conn.execute(update_stmt)
            await conn.commit()

    @threadsafe_method
    async def get_item_async(self, key: Hashable) -> Context:
        stmt = select(self.table.c.context).where(self.table.c.id == str(key))
        async with self.engine.connect() as conn:
            result = await conn.execute(stmt)
            row = result.fetchone()
            if row:
                return Context.cast(row[0])
        raise KeyError

    @threadsafe_method
    async def del_item_async(self, key: Hashable):
        stmt = delete(self.table).where(self.table.c.id == str(key))
        async with self.engine.connect() as conn:
            await conn.execute(stmt)
            await conn.commit()

    @threadsafe_method
    async def contains_async(self, key: Hashable) -> bool:
        stmt = select(self.table.c.context).where(self.table.c.id == str(key))
        async with self.engine.connect() as conn:
            result = await conn.execute(stmt)
            return bool(result.fetchone())

    @threadsafe_method
    async def len_async(self) -> int:
        stmt = select(func.count()).select_from(self.table)
        async with self.engine.connect() as conn:
            result = await conn.execute(stmt)
            return result.fetchone()[0]

    @threadsafe_method
    async def clear_async(self):
        stmt = delete(self.table)
        async with self.engine.connect() as conn:
            await conn.execute(stmt)
            await conn.commit()

    async def _create_self_table(self):
        async with self.engine.begin() as conn:
            if not await conn.run_sync(lambda sync_conn: inspect(sync_conn).has_table(self.table.name)):
                await conn.run_sync(self.table.create, self.engine)

    async def _get_update_stmt(self, insert_stmt):
        if self.dialect == "sqlite":
            return insert_stmt
        elif self.dialect == "mysql":
            update_stmt = insert_stmt.on_duplicate_key_update(context=insert_stmt.inserted.context)
        else:
            update_stmt = insert_stmt.on_conflict_do_update(
                index_elements=["id"], set_=dict(context=insert_stmt.excluded.context)
            )
        return update_stmt

    def _check_availability(self, custom_driver: bool):
        if not custom_driver:
            if self.full_path.startswith("postgresql") and not postgres_available:
                install_suggestion = get_protocol_install_suggestion("postgresql")
                raise ImportError("Packages `sqlalchemy` and/or `asyncpg` are missing.\n" + install_suggestion)
            elif self.full_path.startswith("mysql") and not mysql_available:
                install_suggestion = get_protocol_install_suggestion("mysql")
                raise ImportError("Packages `sqlalchemy` and/or `asyncmy` are missing.\n" + install_suggestion)
            elif self.full_path.startswith("sqlite") and not sqlite_available:
                install_suggestion = get_protocol_install_suggestion("sqlite")
                raise ImportError("Package `sqlalchemy` and/or `aiosqlite` is missing.\n" + install_suggestion)



"""
Command Line Interface
----------------------
This modules defines commands that can be called via the command line interface.

"""

from uuid import uuid4
import tempfile
import shutil
import sys
import argparse
import os
import logging
from urllib import parse
from pathlib import Path
from typing import Optional

try:
    from omegaconf import OmegaConf
    from .utils import get_superset_session, drop_superset_assets
except ImportError:
    raise ImportError("Some packages are not found. Run `pip install chatsky[stats]`")

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

CHATSKY_DIR = Path(__file__).absolute().parent.parent
"""
Root directory of the local `chatsky` installation.

:meta hide-value:
"""
DASHBOARD_DIR = str(CHATSKY_DIR / "config" / "superset_dashboard")
"""
Local path to superset dashboard files to import.

:meta hide-value:
"""
DASHBOARD_SLUG = "chatsky-stats"
"""
This variable stores a slug used for building the http address of the Chatsky dashboard.
"""
DEFAULT_SUPERSET_URL = parse.urlunsplit(("http", "localhost:8088", "/", "", ""))
"""
Default location of the Superset dashboard.
"""

TYPE_MAPPING_CH = {
    "FLOAT": "Nullable(Float64)",
    "STRING": "Nullable(String)",
    "LONGINTEGER": "Nullable(Int64)",
    "INTEGER": "Nullable(Int64)",
    "DATETIME": "Nullable(DateTime)",
}
"""
Mapping of standard sql column types to Clickhouse native types.

:meta hide-value:
"""

CHATSKY_NODE_STATS_STATEMENT = """
WITH main AS (
    SELECT DISTINCT {table}.LogAttributes['context_id'] as context_id,
    toUInt64OrNull({table}.LogAttributes['request_id']) as request_id,
    toDateTime(otel_traces.Timestamp) as start_time,
    otel_traces.SpanName as data_key,
    {table}.Body as data,
    {lblfield} as label,
    {flowfield} as flow_label,
    {nodefield} as node_label,
    {table}.TraceId as trace_id,
    otel_traces.TraceId\nFROM {table}, otel_traces
    WHERE {table}.TraceId = otel_traces.TraceId and data_key = 'get_current_label'
    ORDER BY context_id, request_id
) SELECT context_id,
    request_id,
    start_time,
    data_key,
    data,
    label,
    {label_lag} as prev_label,
    {flow_lag} as prev_flow,
    flow_label,
    node_label
FROM main
"""
CHATSKY_STATS_STATEMENT = """
WITH main AS (
    SELECT DISTINCT {table}.LogAttributes['context_id'] as context_id,
    toUInt64OrNull({table}.LogAttributes['request_id']) as request_id,
    toDateTime(otel_traces.Timestamp) as start_time,
    otel_traces.SpanName as data_key,
    {table}.Body as data,
    {lblfield} as label,
    {flowfield} as flow_label,
    {nodefield} as node_label,
    {table}.TraceId as trace_id,
    otel_traces.TraceId\nFROM {table}, otel_traces
    WHERE {table}.TraceId = otel_traces.TraceId
    ORDER BY data_key, context_id, request_id
) SELECT context_id,
    request_id,
    start_time,
    data_key,
    data,
    label,
    {label_lag} as prev_label,
    {flow_lag} as prev_flow,
    flow_label,
    node_label
FROM main
"""
CHATSKY_FINAL_NODES_STATEMENT = """
WITH main AS (
    SELECT LogAttributes['context_id'] AS context_id,
    max(toUInt64OrNull(LogAttributes['request_id'])) AS max_history
    FROM {table}\nGROUP BY context_id
)
SELECT DISTINCT LogAttributes['context_id'] AS context_id,
toUInt64OrNull({table}.LogAttributes['request_id']) AS request_id,
toDateTime(otel_traces.Timestamp) AS start_time,
{lblfield} AS label,
{flowfield} AS flow_label,
{nodefield} AS node_label
FROM {table}
INNER JOIN main
ON context_id  = main.context_id
AND request_id = main.max_history
INNER JOIN otel_traces
ON {table}.TraceId = otel_traces.TraceId
WHERE otel_traces.SpanName = 'get_current_label'
"""

SQL_STATEMENT_MAPPING = {
    "chatsky_stats.yaml": CHATSKY_STATS_STATEMENT,
    "chatsky_node_stats.yaml": CHATSKY_NODE_STATS_STATEMENT,
    "chatsky_final_nodes.yaml": CHATSKY_FINAL_NODES_STATEMENT,
}
"""
Select statements for dashboard configuration with names and types represented as placeholders.
The placeholder system makes queries database agnostic, required values are set during the import phase.

:meta hide-value:
"""


def import_dashboard(parsed_args: Optional[argparse.Namespace] = None, zip_file: Optional[str] = None):
    """
    Import an Apache Superset dashboard to a local instance with specified arguments.
    Before using the command, make sure you have your Superset instance
    up and running: `ghcr.io/deeppavlov/superset_df_dashboard:latest`.
    The import will override existing dashboard configurations if present.

    :param parsed_args: Command line arguments produced by `argparse`.
    :param zip_file: Zip archived dashboard config.
    """
    host = parsed_args.host if hasattr(parsed_args, "host") else "localhost"
    port = parsed_args.port if hasattr(parsed_args, "port") else "8088"
    superset_url = parse.urlunsplit(("http", f"{host}:{port}", "/", "", ""))
    zip_filename = os.path.basename(zip_file)
    db_password = getattr(parsed_args, "db.password")

    session, headers = get_superset_session(parsed_args, superset_url)
    drop_superset_assets(session, headers, superset_url)
    import_dashboard_url = parse.urljoin(superset_url, "/api/v1/dashboard/import/")
    # upload files
    with open(zip_file, "rb") as f:
        response = session.request(
            "POST",
            import_dashboard_url,
            headers=headers,
            data={
                "passwords": '{"databases/chatsky_database.yaml":"' + db_password + '"}',
                "overwrite": "true",
            },
            files=[("formData", (zip_filename, f, "application/zip"))],
        )
        response.raise_for_status()
        logger.info(f"Upload finished with status {response.status_code}.")


def make_zip_config(parsed_args: argparse.Namespace) -> Path:
    """
    Make a zip-archived Apache Superset dashboard config, using specified arguments.

    :param parsed_args: Command line arguments produced by `argparse`.
    """
    if hasattr(parsed_args, "outfile") and parsed_args.outfile:
        outfile_name = parsed_args.outfile
    else:
        outfile_name = f"config_{str(uuid4())}.zip"

    file_conf = OmegaConf.load(parsed_args.file)
    sys.argv = [__file__] + [f"{key}={value}" for key, value in parsed_args.__dict__.items() if value]
    cmd_conf = OmegaConf.from_cli()
    cli_conf = OmegaConf.merge(file_conf, cmd_conf)

    if OmegaConf.select(cli_conf, "db.driver") == "clickhousedb+connect":
        params = dict(
            table="${db.table}",
            label_lag="lagInFrame(label) OVER "
            "(PARTITION BY context_id ORDER BY request_id ASC "
            "ROWS BETWEEN UNBOUNDED PRECEDING AND UNBOUNDED FOLLOWING)",
            flow_lag="lagInFrame(flow_label) OVER "
            "(PARTITION BY context_id ORDER BY request_id ASC "
            "ROWS BETWEEN UNBOUNDED PRECEDING AND UNBOUNDED FOLLOWING)",
            texttype="String",
            lblfield="JSON_VALUE(${db.table}.Body, '$.label')",
            flowfield="JSON_VALUE(${db.table}.Body, '$.flow')",
            nodefield="JSON_VALUE(${db.table}.Body, '$.node')",
        )
    else:
        raise ValueError("The only supported database driver is 'clickhousedb+connect'.")

    conf = SQL_STATEMENT_MAPPING.copy()
    for key in conf.keys():
        conf[key] = {}
        conf[key]["sql"] = SQL_STATEMENT_MAPPING[key].format(**params)

    resolve_conf = OmegaConf.create(
        {
            "database": {
                "sqlalchemy_uri": "${db.driver}://${db.user}:XXXXXXXXXX@${db.host}:${db.port}/${db.name}",
            },
            **conf,
        }
    )

    user_config = OmegaConf.merge(cli_conf, resolve_conf)
    OmegaConf.resolve(user_config)

    with tempfile.TemporaryDirectory() as temp_config_dir:
        nested_temp_dir = os.path.join(temp_config_dir, "superset_dashboard")
        logger.info(f"Copying config files to temporary directory: {nested_temp_dir}.")

        shutil.copytree(DASHBOARD_DIR, nested_temp_dir)
        database_dir = Path(os.path.join(nested_temp_dir, "databases"))
        dataset_dir = Path(os.path.join(nested_temp_dir, "datasets/chatsky_database"))

        logger.info("Overriding the initial configuration.")
        # overwrite sqlalchemy uri
        for filepath in database_dir.iterdir():
            file_config = OmegaConf.load(filepath)
            new_file_config = OmegaConf.merge(file_config, OmegaConf.select(user_config, "database"))
            OmegaConf.save(new_file_config, filepath)

        # overwrite sql expressions and column types
        for filepath in dataset_dir.iterdir():
            file_config = OmegaConf.load(filepath)
            new_file_config = OmegaConf.merge(file_config, getattr(user_config, filepath.name))
            if OmegaConf.select(cli_conf, "db.driver") == "clickhousedb+connect":
                for col in OmegaConf.select(new_file_config, "columns"):
                    col.type = TYPE_MAPPING_CH.get(col.type, col.type)
            OmegaConf.save(new_file_config, filepath)

        if ".zip" not in outfile_name:
            raise ValueError(f"Outfile name missing .zip extension: {outfile_name}.")
        logger.info(f"Saving the archive to {outfile_name}.")
        shutil.make_archive(outfile_name[: outfile_name.rindex(".zip")], format="zip", root_dir=temp_config_dir)

    return Path(outfile_name)



"""
Conditions
----------
Conditions are one of the most important components of the dialog graph.
They determine the possibility of transition from one node of the graph to another.
The conditions are used to specify when a particular transition should occur, based on certain criteria.
This module contains a standard set of scripting conditions that can be used to control the flow of a conversation.
These conditions can be used to check the current context, the user's input,
or other factors that may affect the conversation flow.
"""

from typing import Callable, Pattern, Union, List, Optional
import logging
import re

from pydantic import validate_call

from chatsky.pipeline import Pipeline
from chatsky.script import NodeLabel2Type, Context, Message
from chatsky.script.core.message import CallbackQuery

logger = logging.getLogger(__name__)


@validate_call
def exact_match(match: Union[str, Message], skip_none: bool = True) -> Callable[[Context, Pipeline], bool]:
    """
    Return function handler. This handler returns `True` only if the last user phrase
    is the same `Message` as the `match`.
    If `skip_none` the handler will not compare `None` fields of `match`.

    :param match: A `Message` variable to compare user request with.
        Can also accept `str`, which will be converted into a `Message` with its text field equal to `match`.
    :param skip_none: Whether fields should be compared if they are `None` in :py:const:`match`.
    """

    def exact_match_condition_handler(ctx: Context, pipeline: Pipeline) -> bool:
        request = ctx.last_request
        nonlocal match
        if isinstance(match, str):
            match = Message(text=match)
        if request is None:
            return False
        for field in match.model_fields:
            match_value = match.__getattribute__(field)
            if skip_none and match_value is None:
                continue
            if field in request.model_fields.keys():
                if request.__getattribute__(field) != match.__getattribute__(field):
                    return False
            else:
                return False
        return True

    return exact_match_condition_handler


@validate_call
def has_text(text: str) -> Callable[[Context, Pipeline], bool]:
    """
    Return function handler. This handler returns `True` only if the last user phrase
    contains the phrase specified in `text`.

    :param text: A `str` variable to look for within the user request.
    """

    def has_text_condition_handler(ctx: Context, pipeline: Pipeline) -> bool:
        request = ctx.last_request
        return text in request.text

    return has_text_condition_handler


@validate_call
def regexp(pattern: Union[str, Pattern], flags: Union[int, re.RegexFlag] = 0) -> Callable[[Context, Pipeline], bool]:
    """
    Return function handler. This handler returns `True` only if the last user phrase contains
    `pattern` with `flags`.

    :param pattern: The `RegExp` pattern.
    :param flags: Flags for this pattern. Defaults to 0.
    """
    pattern = re.compile(pattern, flags)

    def regexp_condition_handler(ctx: Context, pipeline: Pipeline) -> bool:
        request = ctx.last_request
        if isinstance(request, Message):
            if request.text is None:
                return False
            return bool(pattern.search(request.text))
        else:
            logger.error(f"request has to be str type, but got request={request}")
            return False

    return regexp_condition_handler


@validate_call
def check_cond_seq(cond_seq: list):
    """
    Check if the list consists only of Callables.

    :param cond_seq: List of conditions to check.
    """
    for cond in cond_seq:
        if not callable(cond):
            raise TypeError(f"{cond_seq} has to consist of callable objects")


_any = any
"""
_any is an alias for any.
"""
_all = all
"""
_all is an alias for all.
"""


@validate_call
def aggregate(cond_seq: list, aggregate_func: Callable = _any) -> Callable[[Context, Pipeline], bool]:
    """
    Aggregate multiple functions into one by using aggregating function.

    :param cond_seq: List of conditions to check.
    :param aggregate_func: Function to aggregate conditions. Defaults to :py:func:`_any`.
    """
    check_cond_seq(cond_seq)

    def aggregate_condition_handler(ctx: Context, pipeline: Pipeline) -> bool:
        try:
            return bool(aggregate_func([cond(ctx, pipeline) for cond in cond_seq]))
        except Exception as exc:
            logger.error(f"Exception {exc} for {cond_seq}, {aggregate_func} and {ctx.last_request}", exc_info=exc)
            return False

    return aggregate_condition_handler


@validate_call
def any(cond_seq: list) -> Callable[[Context, Pipeline], bool]:
    """
    Return function handler. This handler returns `True`
    if any function from the list is `True`.

    :param cond_seq: List of conditions to check.
    """
    _agg = aggregate(cond_seq, _any)

    def any_condition_handler(ctx: Context, pipeline: Pipeline) -> bool:
        return _agg(ctx, pipeline)

    return any_condition_handler


@validate_call
def all(cond_seq: list) -> Callable[[Context, Pipeline], bool]:
    """
    Return function handler. This handler returns `True` only
    if all functions from the list are `True`.

    :param cond_seq: List of conditions to check.
    """
    _agg = aggregate(cond_seq, _all)

    def all_condition_handler(ctx: Context, pipeline: Pipeline) -> bool:
        return _agg(ctx, pipeline)

    return all_condition_handler


@validate_call
def negation(condition: Callable) -> Callable[[Context, Pipeline], bool]:
    """
    Return function handler. This handler returns negation of the :py:func:`~condition`: `False`
    if :py:func:`~condition` holds `True` and returns `True` otherwise.

    :param condition: Any :py:func:`~condition`.
    """

    def negation_condition_handler(ctx: Context, pipeline: Pipeline) -> bool:
        return not condition(ctx, pipeline)

    return negation_condition_handler


@validate_call
def has_last_labels(
    flow_labels: Optional[List[str]] = None,
    labels: Optional[List[NodeLabel2Type]] = None,
    last_n_indices: int = 1,
) -> Callable[[Context, Pipeline], bool]:
    """
    Return condition handler. This handler returns `True` if any label from
    last `last_n_indices` context labels is in
    the `flow_labels` list or in
    the `labels` list.

    :param flow_labels: List of labels to check. Every label has type `str`. Empty if not set.
    :param labels: List of labels corresponding to the nodes. Empty if not set.
    :param last_n_indices: Number of last utterances to check.
    """
    # todo: rewrite docs & function itself
    flow_labels = [] if flow_labels is None else flow_labels
    labels = [] if labels is None else labels

    def has_last_labels_condition_handler(ctx: Context, pipeline: Pipeline) -> bool:
        label = list(ctx.labels.values())[-last_n_indices:]
        for label in list(ctx.labels.values())[-last_n_indices:]:
            label = label if label else (None, None)
            if label[0] in flow_labels or label in labels:
                return True
        return False

    return has_last_labels_condition_handler


@validate_call
def true() -> Callable[[Context, Pipeline], bool]:
    """
    Return function handler. This handler always returns `True`.
    """

    def true_handler(ctx: Context, pipeline: Pipeline) -> bool:
        return True

    return true_handler


@validate_call
def false() -> Callable[[Context, Pipeline], bool]:
    """
    Return function handler. This handler always returns `False`.
    """

    def false_handler(ctx: Context, pipeline: Pipeline) -> bool:
        return False

    return false_handler


# aliases
agg = aggregate
"""
:py:func:`~agg` is an alias for :py:func:`~aggregate`.
"""
neg = negation
"""
:py:func:`~neg` is an alias for :py:func:`~negation`.
"""


def has_callback_query(expected_query_string: str) -> Callable[[Context, Pipeline], bool]:
    """
    Condition that checks if :py:attr:`~.CallbackQuery.query_string`
    of the last message matches `expected_query_string`.

    :param expected_query_string: The expected query string to compare with.
    :return: The callback query comparator function.
    """

    def has_callback_query_handler(ctx: Context, _: Pipeline) -> bool:
        last_request = ctx.last_request
        if last_request is None or last_request.attachments is None:
            return False
        return CallbackQuery(query_string=expected_query_string) in last_request.attachments

    return has_callback_query_handler



"""
Telegram Interfaces
-------------------
This module provides concrete implementations of the
:py:class:`~._AbstractTelegramInterface`.
"""

from pathlib import Path
from typing import Any, Optional

from chatsky.pipeline.types import PipelineRunnerFunction

from .abstract import _AbstractTelegramInterface

try:
    from telegram import Update
except ImportError:
    Update = Any


class LongpollingInterface(_AbstractTelegramInterface):
    """
    Telegram messenger interface, that requests Telegram API in a loop.

    :param token: The Telegram bot token.
    :param attachments_directory: The directory for storing attachments.
    :param interval: A time interval between polls (in seconds).
    :param timeout: Timeout in seconds for long polling.
    """

    def __init__(
        self, token: str, attachments_directory: Optional[Path] = None, interval: int = 2, timeout: int = 20
    ) -> None:
        super().__init__(token, attachments_directory)
        self.interval = interval
        self.timeout = timeout

    async def connect(self, pipeline_runner: PipelineRunnerFunction, *args, **kwargs):
        await super().connect(pipeline_runner, *args, **kwargs)
        self.application.run_polling(
            poll_interval=self.interval, timeout=self.timeout, allowed_updates=Update.ALL_TYPES
        )


class WebhookInterface(_AbstractTelegramInterface):
    """
    Telegram messenger interface, that brings a special webserver up
    and registers up for listening for Telegram updates.

    :param token: The Telegram bot token.
    :param attachments_directory: The directory for storing attachments.
    :param host: Local host name (or IP address).
    :param port: Local port for running Telegram webhook.
    """

    def __init__(
        self, token: str, attachments_directory: Optional[Path] = None, host: str = "localhost", port: int = 844
    ):
        super().__init__(token, attachments_directory)
        self.listen = host
        self.port = port

    async def connect(self, pipeline_runner: PipelineRunnerFunction, *args, **kwargs):
        await super().connect(pipeline_runner, *args, **kwargs)
        self.application.run_webhook(listen=self.listen, port=self.port, allowed_updates=Update.ALL_TYPES)



# flake8: noqa: F401

from chatsky.pipeline import Pipeline
from chatsky.pipeline.types import ExtraHandlerRuntimeInfo
from chatsky.script import Context, Script

Script.model_rebuild()
Context.model_rebuild()
ExtraHandlerRuntimeInfo.model_rebuild()
